<?php

declare(strict_types=1);

namespace Dedi\SyliusImportExportPlugin\ShippingMethod\Importer\Processor;

use Dedi\SyliusImportExportPlugin\ShippingMethod\Exporter\ShippingMethodDataProviderInterface;
use Dedi\SyliusImportExportPlugin\ShippingMethod\ShippingMethodRuleTrait;
use Doctrine\ORM\EntityManagerInterface;
use FriendsOfSylius\SyliusImportExportPlugin\Exception\ImporterException;
use FriendsOfSylius\SyliusImportExportPlugin\Processor\MetadataValidatorInterface;
use FriendsOfSylius\SyliusImportExportPlugin\Processor\ResourceProcessorInterface;
use Sylius\Component\Addressing\Model\ZoneInterface;
use Sylius\Component\Channel\Repository\ChannelRepositoryInterface;
use Sylius\Component\Core\Model\ChannelInterface;
use Sylius\Component\Core\Model\ShippingMethodInterface;
use Sylius\Component\Core\Repository\ShippingMethodRepositoryInterface;
use Sylius\Component\Core\Shipping\Checker\Rule\OrderTotalLessThanOrEqualRuleChecker;
use Sylius\Component\Resource\Factory\FactoryInterface;
use Sylius\Component\Resource\Repository\RepositoryInterface;
use Sylius\Component\Shipping\Calculator\DefaultCalculators;
use Sylius\Component\Shipping\Checker\Rule\TotalWeightGreaterThanOrEqualRuleChecker;
use Sylius\Component\Shipping\Checker\Rule\TotalWeightLessThanOrEqualRuleChecker;
use Sylius\Component\Taxation\Model\TaxRateInterface;
use Sylius\Component\Taxation\Repository\TaxCategoryRepositoryInterface;
use Sylius\Component\Taxation\Resolver\TaxRateResolverInterface;

class ShippingMethodProcessor implements ResourceProcessorInterface
{
    use ShippingMethodRuleTrait;

    private readonly MetadataValidatorInterface $metadataValidator;

    private readonly EntityManagerInterface $entityManager;

    private readonly ShippingMethodRepositoryInterface $shippingMethodRepository;

    private readonly ShippingMethodDataProviderInterface $shippingMethodDataProvider;

    private readonly FactoryInterface $factory;

    private readonly RepositoryInterface $zoneRepository;

    private readonly ChannelRepositoryInterface $channelRepository;

    private readonly TaxRateResolverInterface $taxRateResolver;

    private readonly TaxCategoryRepositoryInterface $taxCategoryRepository;

    public function __construct(
        MetadataValidatorInterface $metadataValidator,
        EntityManagerInterface $entityManager,
        ShippingMethodRepositoryInterface $shippingMethodRepository,
        ShippingMethodDataProviderInterface $shippingMethodDataProvider,
        RepositoryInterface $zoneRepository,
        ChannelRepositoryInterface $channelRepository,
        FactoryInterface $factory,
        TaxRateResolverInterface $taxRateResolver,
        TaxCategoryRepositoryInterface $taxCategoryRepository,
    ) {
        $this->metadataValidator = $metadataValidator;
        $this->entityManager = $entityManager;
        $this->shippingMethodRepository = $shippingMethodRepository;
        $this->shippingMethodDataProvider = $shippingMethodDataProvider;
        $this->zoneRepository = $zoneRepository;
        $this->channelRepository = $channelRepository;
        $this->factory = $factory;
        $this->taxRateResolver = $taxRateResolver;
        $this->taxCategoryRepository = $taxCategoryRepository;
    }

    /**
     * @throws ImporterException
     */
    public function process(array $data): void
    {
        $headers = $this->getHeaders();

        $this->metadataValidator->validateHeaders($headers, $data);

        $data = array_combine(array_keys($headers), $data);

        $mandatoryData = $this->getMandatoryData();
        foreach ($mandatoryData as $key) {
            if ('' === (string) $data[$key]) {
                throw new ImporterException(
                    sprintf(
                        'One of the mandatory parameters of the following list is empty : %s',
                        implode(',', $mandatoryData),
                    ),
                );
            }
        }

        $shippingMethod = $this->getShippingMethod($data);

        $this->entityManager->persist($shippingMethod);
    }

    private function getHeaders(): array
    {
        return $this->shippingMethodDataProvider->getHeaders();
    }

    protected function getMandatoryData(): array
    {
        return [
            'code',
            'channel',
            'price_including_tax',
        ];
    }

    protected function getShippingMethod(array $data): ShippingMethodInterface
    {
        /** @var ?ShippingMethodInterface $shippingMethod */
        $shippingMethod = $this->shippingMethodRepository->findOneBy(['code' => $data['code']]);
        if (null === $shippingMethod) {
            $defaultTaxCategory = $this->taxCategoryRepository->findOneBy([]);

            /** @var ShippingMethodInterface $shippingMethod */
            $shippingMethod = $this->factory->createNew();
            $shippingMethod->setCode((string) $data['code']);
            $shippingMethod->setCalculator(DefaultCalculators::FLAT_RATE);
            $shippingMethod->setTaxCategory($defaultTaxCategory);
        }

        $shippingMethod->setName((string) $data['name']);
        $shippingMethod->setDescription((string) $data['description']);
        $shippingMethod->setPosition((int) $data['position']);

        /** @var ZoneInterface|null $zone */
        $zone = $this->zoneRepository->findOneBy(['code' => $data['zone']]);
        $shippingMethod->setZone($zone);

        foreach ($shippingMethod->getChannels() as $channel) {
            $shippingMethod->removeChannel($channel);
        }

        $channels = [];
        foreach (explode('|', (string) $data['channel']) as $channelCode) {
            $channelCode = \trim($channelCode);

            /** @var ChannelInterface|null $channel */
            $channel = $this->channelRepository->findOneBy(['code' => $channelCode]);
            if (null === $channel) {
                throw new ImporterException(sprintf('Channel with code %s does not exist', $channelCode));
            }

            $shippingMethod->addChannel($channel);
            $channels[] = $channel;
        }

        if ('' !== (string) $data['min_weight']) {
            $this->addRule(
                $shippingMethod,
                TotalWeightGreaterThanOrEqualRuleChecker::TYPE,
                ['weight' => (int) $data['min_weight']],
            );
        } else {
            $rule = $this->getRule($shippingMethod, TotalWeightGreaterThanOrEqualRuleChecker::TYPE);
            if ($rule !== null) {
                $shippingMethod->removeRule($rule);
            }
        }

        if ('' !== (string) $data['max_weight']) {
            $this->addRule(
                $shippingMethod,
                TotalWeightLessThanOrEqualRuleChecker::TYPE,
                ['weight' => (int) $data['max_weight']],
            );
        } else {
            $rule = $this->getRule($shippingMethod, TotalWeightLessThanOrEqualRuleChecker::TYPE);
            if ($rule !== null) {
                $shippingMethod->removeRule($rule);
            }
        }

        $rule = $this->getRule($shippingMethod, OrderTotalLessThanOrEqualRuleChecker::TYPE);
        if ($rule !== null) {
            $shippingMethod->removeRule($rule);
        }
        $taxRate = $this->taxRateResolver->resolve($shippingMethod);

        $maxCartAmountsWithoutTax = explode('|', (string) $data['max_cart_amount_excluding_tax']);
        if ('' !== (string) $data['max_cart_amount_excluding_tax']) {
            if (\count($maxCartAmountsWithoutTax) === 1) {
                $maxCartAmountsWithoutTax = array_fill(0, \count($channels), current($maxCartAmountsWithoutTax));
            } elseif (\count($maxCartAmountsWithoutTax) !== \count($channels)) {
                throw new ImporterException('File must contain one max cart amount value per channel');
            }
        }

        foreach ($maxCartAmountsWithoutTax as $key => $maxCartAmountWithoutTax) {
            $maxCartAmountWithoutTax = \trim($maxCartAmountWithoutTax);
            if ('' === $maxCartAmountWithoutTax) {
                continue;
            }
            $maxCartAmountWithTax = $this->getAmountWithTax((float) $maxCartAmountWithoutTax * 100, $taxRate);
            $this->addRule(
                $shippingMethod,
                OrderTotalLessThanOrEqualRuleChecker::TYPE,
                [
                    $channels[$key]->getCode() => [
                        'amount' => (int) $maxCartAmountWithTax,
                    ],
                ],
            );
        }

        $pricesIncludingTax = explode('|', (string) $data['price_including_tax']);
        if (\count($pricesIncludingTax) === 1) {
            $pricesIncludingTax = array_fill(0, \count($channels), current($pricesIncludingTax));
        } elseif (\count($pricesIncludingTax) !== \count($channels)) {
            throw new ImporterException('File must contain one shipping amount value per channel');
        }

        foreach ($pricesIncludingTax as $key => $priceIncludingTax) {
            $priceIncludingTax = \trim($priceIncludingTax);
            if ('' === $priceIncludingTax) {
                continue;
            }
            $shippingMethod->setConfiguration(array_replace_recursive(
                $shippingMethod->getConfiguration(),
                [
                    $channels[$key]->getCode() => [
                        'amount' => (int) ((float) $priceIncludingTax * 100),
                    ],
                ],
            ));
        }

        return $shippingMethod;
    }

    private function getAmountWithTax(float $amountWithoutTax, ?TaxRateInterface $taxRate): float
    {
        if (null === $taxRate || false === $taxRate->isIncludedInPrice()) {
            return $amountWithoutTax;
        }

        $taxAmount = round($amountWithoutTax * $taxRate->getAmount());

        return $amountWithoutTax + $taxAmount;
    }
}
